@@ -14,17 +14,20 @@
 
 module Language.Haskell.Meta.Syntax.Translate (
     module Language.Haskell.Meta.Syntax.Translate
+  , TyVarBndr_
 ) where
 
-import qualified Data.Char                    as Char
-import qualified Data.List                    as List
-import qualified Language.Haskell.Exts.SrcLoc as Exts.SrcLoc
-import qualified Language.Haskell.Exts.Syntax as Exts
-import qualified Language.Haskell.TH.Syntax   as TH
+import qualified Data.Char                      as Char
+import qualified Data.List                      as List
+import qualified Language.Haskell.Exts.SrcLoc   as Exts.SrcLoc
+import qualified Language.Haskell.Exts.Syntax   as Exts
+import           Language.Haskell.Meta.THCompat (TyVarBndr_)
+import qualified Language.Haskell.Meta.THCompat as Compat
+import qualified Language.Haskell.TH.Lib        as TH
+import qualified Language.Haskell.TH.Syntax     as TH
 
 -----------------------------------------------------------------------------
 
-
 class ToName a where toName :: a -> TH.Name
 class ToNames a where toNames :: a -> [TH.Name]
 class ToLit  a where toLit  :: a -> TH.Lit
@@ -37,19 +40,11 @@ class ToStmt a where toStmt :: a -> TH.Stmt
 class ToLoc  a where toLoc  :: a -> TH.Loc
 class ToCxt  a where toCxt  :: a -> TH.Cxt
 class ToPred a where toPred :: a -> TH.Pred
-class ToTyVars a where toTyVars :: a -> [TH.TyVarBndr]
+class ToTyVars a where toTyVars :: a -> [TyVarBndr_ ()]
 class ToMaybeKind a where toMaybeKind :: a -> Maybe TH.Kind
-#if MIN_VERSION_template_haskell(2,11,0)
 class ToInjectivityAnn a where toInjectivityAnn :: a -> TH.InjectivityAnn
-#endif
 
-#if MIN_VERSION_template_haskell(2,12,0)
 type DerivClause = TH.DerivClause
-#elif MIN_VERSION_template_haskell(2,11,0)
-type DerivClause = TH.Pred
-#else
-type DerivClause = TH.Name
-#endif
 
 class ToDerivClauses a where toDerivClauses :: a -> [DerivClause]
 
@@ -202,7 +197,6 @@ instance ToLit (Exts.Literal l) where
 
 -- * ToPat HsPat
 
-
 instance ToPat (Exts.Pat l) where
   toPat (Exts.PVar _ n)
     = TH.VarP (toName n)
@@ -216,7 +210,7 @@ instance ToPat (Exts.Pat l) where
     TH.DoublePrimL r'' -> TH.DoublePrimL (negate r'')
     _                  -> nonsense "toPat" "negating wrong kind of literal" l
   toPat (Exts.PInfixApp _ p n q) = TH.UInfixP (toPat p) (toName n) (toPat q)
-  toPat (Exts.PApp _ n ps) = TH.ConP (toName n) (fmap toPat ps)
+  toPat (Exts.PApp _ n ps) = Compat.conP (toName n) (fmap toPat ps)
   toPat (Exts.PTuple _ Exts.Boxed ps) = TH.TupP (fmap toPat ps)
   toPat (Exts.PTuple _ Exts.Unboxed ps) = TH.UnboxedTupP (fmap toPat ps)
   toPat (Exts.PList _ ps) = TH.ListP (fmap toPat ps)
@@ -266,12 +260,11 @@ instance ToExp (Exts.Exp l) where
   toExp e@Exts.IPVar{}                 = noTH "toExp" e
   toExp (Exts.Con _ n)                 = TH.ConE (toName n)
   toExp (Exts.Lit _ l)                 = TH.LitE (toLit l)
-  toExp (Exts.InfixApp _ e o f)        = TH.UInfixE (toExp e) (toExp o) (toExp f)
-#if MIN_VERSION_template_haskell(2,12,0)
-  toExp (Exts.App _ e (Exts.TypeApp _ t)) = TH.AppTypeE (toExp e) (toType t)
-#else
-  toExp (Exts.App _ _ e@Exts.TypeApp{}) = noTHyet "toExp" "2.12.0" e
+#if MIN_VERSION_template_haskell(2,13,0)
+  toExp (Exts.OverloadedLabel _ s)     = TH.LabelE s
 #endif
+  toExp (Exts.InfixApp _ e o f)        = TH.UInfixE (toExp e) (toExp o) (toExp f)
+  toExp (Exts.App _ e (Exts.TypeApp _ t)) = TH.AppTypeE (toExp e) (toType t)
   toExp (Exts.App _ e f)               = TH.AppE (toExp e) (toExp f)
   toExp (Exts.NegApp _ e)              = TH.AppE (TH.VarE 'negate) (toExp e)
   toExp (Exts.Lambda _ ps e)           = TH.LamE (fmap toPat ps) (toExp e)
@@ -279,7 +272,11 @@ instance ToExp (Exts.Exp l) where
   toExp (Exts.If _ a b c)              = TH.CondE (toExp a) (toExp b) (toExp c)
   toExp (Exts.MultiIf _ ifs)           = TH.MultiIfE (map toGuard ifs)
   toExp (Exts.Case _ e alts)           = TH.CaseE (toExp e) (map toMatch alts)
+#if MIN_VERSION_template_haskell(2,17,0)
+  toExp (Exts.Do _ ss)                 = TH.DoE Nothing (map toStmt ss)
+#else
   toExp (Exts.Do _ ss)                 = TH.DoE (map toStmt ss)
+#endif
   toExp e@Exts.MDo{}                   = noTH "toExp" e
   toExp (Exts.Tuple _ Exts.Boxed xs)   = TH.TupE (fmap toTupEl xs)
   toExp (Exts.Tuple _ Exts.Unboxed xs) = TH.UnboxedTupE (fmap toTupEl xs)
@@ -347,9 +344,14 @@ instance ToName (Exts.TyVarBind l) where
 instance ToName TH.Name where
   toName = id
 
-instance ToName TH.TyVarBndr where
-  toName (TH.PlainTV n)    = n
-  toName (TH.KindedTV n _) = n
+instance ToName (Compat.TyVarBndr_ flag) where
+#if MIN_VERSION_template_haskell(2,17,0)
+  toName (TH.PlainTV n _)    = n
+  toName (TH.KindedTV n _ _) = n
+#else
+  toName (TH.PlainTV n)      = n
+  toName (TH.KindedTV n _)   = n
+#endif
 
 #if !MIN_VERSION_haskell_src_exts(1,21,0)
 instance ToType (Exts.Kind l) where
@@ -375,12 +377,26 @@ instance ToType (Exts.Kind l) where
 toKind :: Exts.Kind l -> TH.Kind
 toKind = toType
 
-toTyVar :: Exts.TyVarBind l -> TH.TyVarBndr
+toTyVar :: Exts.TyVarBind l -> TyVarBndr_ ()
+#if MIN_VERSION_template_haskell(2,17,0)
+toTyVar (Exts.KindedVar _ n k) = TH.KindedTV (toName n) () (toKind k)
+toTyVar (Exts.UnkindedVar _ n) = TH.PlainTV (toName n) ()
+#else
 toTyVar (Exts.KindedVar _ n k) = TH.KindedTV (toName n) (toKind k)
 toTyVar (Exts.UnkindedVar _ n) = TH.PlainTV (toName n)
+#endif
+
+#if MIN_VERSION_template_haskell(2,17,0)
+toTyVarSpec :: TyVarBndr_ () -> TH.TyVarBndrSpec
+toTyVarSpec (TH.KindedTV n () k) = TH.KindedTV n TH.SpecifiedSpec k
+toTyVarSpec (TH.PlainTV n ())    = TH.PlainTV n TH.SpecifiedSpec
+#else
+toTyVarSpec :: TyVarBndr_ flag -> TyVarBndr_ flag
+toTyVarSpec = id
+#endif
 
 instance ToType (Exts.Type l) where
-  toType (Exts.TyForall _ tvbM cxt t) = TH.ForallT (maybe [] (fmap toTyVar) tvbM) (toCxt cxt) (toType t)
+  toType (Exts.TyForall _ tvbM cxt t) = TH.ForallT (maybe [] (fmap (toTyVarSpec . toTyVar)) tvbM) (toCxt cxt) (toType t)
   toType (Exts.TyFun _ a b) = toType a .->. toType b
   toType (Exts.TyList _ t) = TH.ListT `TH.AppT` toType t
   toType (Exts.TyTuple _ b ts) = foldAppT (tuple . length $ ts) (fmap toType ts)
@@ -408,7 +424,7 @@ instance ToType (Exts.Type l) where
     Exts.PromotedString _ _ s -> TH.LitT $ TH.StrTyLit s
     Exts.PromotedCon _ _q n -> TH.PromotedT $ toName n
     Exts.PromotedList _ _q ts -> foldr (\t pl -> TH.PromotedConsT `TH.AppT` toType t `TH.AppT` pl) TH.PromotedNilT ts
-    Exts.PromotedTuple _ ts -> foldr (\t pt -> pt `TH.AppT` toType t) (TH.PromotedTupleT $ length ts) ts
+    Exts.PromotedTuple _ ts -> foldl (\pt t -> pt `TH.AppT` toType t) (TH.PromotedTupleT $ length ts) ts
     Exts.PromotedUnit _ -> TH.PromotedT ''()
   toType (Exts.TyEquals _ t1 t2) = TH.EqualityT `TH.AppT` toType t1 `TH.AppT` toType t2
   toType t@Exts.TySplice{} = noTH "toType" t
@@ -422,7 +438,6 @@ instance ToType (Exts.Type l) where
   -- toType (Exts.TyInfix _ _ (Exts.PromotedName _ _) _)
 
 toStrictType :: Exts.Type l -> TH.StrictType
-#if MIN_VERSION_template_haskell(2,11,0)
 toStrictType (Exts.TyBang _ s u t) = (TH.Bang (toUnpack u) (toStrict s), toType t)
     where
       toStrict (Exts.LazyTy _)        = TH.SourceLazy
@@ -432,19 +447,6 @@ toStrictType (Exts.TyBang _ s u t) = (TH.Bang (toUnpack u) (toStrict s), toType
       toUnpack (Exts.NoUnpack _)       = TH.SourceNoUnpack
       toUnpack (Exts.NoUnpackPragma _) = TH.NoSourceUnpackedness
 toStrictType x = (TH.Bang TH.NoSourceUnpackedness TH.NoSourceStrictness, toType x)
-#else
--- TODO: what is this comment? Outdated?
--- TyBang l (BangType l) (Unpackedness l) (Type l)
--- data BangType l = BangedTy l        | LazyTy l | NoStrictAnnot l
--- data Unpackedness l = Unpack l | NoUnpack l | NoUnpackPragma l
-toStrictType (Exts.TyBang _ b u t) = (toStrict b u, toType t)
-    where
-      toStrict :: Exts.BangType l -> Exts.Unpackedness l -> TH.Strict
-      toStrict (Exts.BangedTy _) _ = TH.IsStrict
-      toStrict _ (Exts.Unpack _)   = TH.Unpacked
-      toStrict _ _                 = TH.NotStrict
-toStrictType x = (TH.NotStrict, toType x)
-#endif
 
 (.->.) :: TH.Type -> TH.Type -> TH.Type
 a .->. b = TH.AppT (TH.AppT TH.ArrowT a) b
@@ -467,26 +469,11 @@ instance ToPred (Exts.Asst l) where
     -- toPred p = todo "toPred" p
 
 instance ToDerivClauses (Exts.Deriving l) where
-#if MIN_VERSION_template_haskell(2,12,0)
 #if MIN_VERSION_haskell_src_exts(1,20,0)
   toDerivClauses (Exts.Deriving _ strat irules) = [TH.DerivClause (fmap toDerivStrategy strat) (map toType irules)]
 #else
   toDerivClauses (Exts.Deriving _ irules) = [TH.DerivClause Nothing (map toType irules)]
 #endif
-#elif MIN_VERSION_template_haskell(2,11,0)
-#if MIN_VERSION_haskell_src_exts(1,20,0)
-  toDerivClauses (Exts.Deriving _ _ irules) = map toType irules
-#else
-  toDerivClauses (Exts.Deriving _ irules) = map toType irules
-#endif
-#else
--- template-haskell < 2.11
-#if MIN_VERSION_haskell_src_exts(1,20,0)
-  toDerivClauses (Exts.Deriving _ _ irules) = concatMap toNames irules
-#else
-  toDerivClauses (Exts.Deriving _ irules) = concatMap toNames irules
-#endif
-#endif
 
 instance ToDerivClauses a => ToDerivClauses (Maybe a) where
   toDerivClauses Nothing  = []
@@ -496,20 +483,15 @@ instance ToDerivClauses a => ToDerivClauses [a] where
   toDerivClauses = concatMap toDerivClauses
 
 
-#if MIN_VERSION_template_haskell(2,12,0) && MIN_VERSION_haskell_src_exts(1,20,0)
 toDerivStrategy :: (Exts.DerivStrategy l) -> TH.DerivStrategy
 toDerivStrategy (Exts.DerivStock _)    = TH.StockStrategy
 toDerivStrategy (Exts.DerivAnyclass _) = TH.AnyclassStrategy
 toDerivStrategy (Exts.DerivNewtype _)  = TH.NewtypeStrategy
-#if MIN_VERSION_haskell_src_exts(1,21,0)
-#if MIN_VERSION_template_haskell(2,14,0)
+#if MIN_VERSION_haskell_src_exts(1,21,0) && MIN_VERSION_template_haskell(2,14,0)
 toDerivStrategy (Exts.DerivVia _ t)    = TH.ViaStrategy (toType t)
 #else
 toDerivStrategy d@Exts.DerivVia{}      = noTHyet "toDerivStrategy" "2.14" d
 #endif
-#endif
-
-#endif
 
 
 -- TODO LumiGuide
@@ -555,9 +537,7 @@ instance ToDec (Exts.Decl l) where
         Exts.DataType _ -> TH.DataD (toCxt cxt)
                              (toName h)
                              (toTyVars h)
-#if MIN_VERSION_template_haskell(2,11,0)
                              Nothing
-#endif
                              (fmap qualConDeclToCon qcds)
                              (toDerivClauses qns)
         Exts.NewType _  -> let qcd = case qcds of
@@ -567,9 +547,7 @@ instance ToDec (Exts.Decl l) where
                         in TH.NewtypeD (toCxt cxt)
                                     (toName h)
                                     (toTyVars h)
-#if MIN_VERSION_template_haskell(2,11,0)
                                     Nothing
-#endif
                                     (qualConDeclToCon qcd)
                                     (toDerivClauses qns)
 
@@ -587,7 +565,6 @@ instance ToDec (Exts.Decl l) where
    where
     inline | b = TH.Inline | otherwise = TH.NoInline
 
-#if MIN_VERSION_template_haskell(2,11,0)
   toDec (Exts.TypeFamDecl _ h sig inj)
     = TH.OpenTypeFamilyD $ TH.TypeFamilyHead (toName h)
                                        (toTyVars h)
@@ -595,12 +572,6 @@ instance ToDec (Exts.Decl l) where
                                        (fmap toInjectivityAnn inj)
   toDec (Exts.DataFamDecl _ _ h sig)
     = TH.DataFamilyD (toName h) (toTyVars h) (toMaybeKind sig)
-#else
-  toDec (Exts.TypeFamDecl _ h sig inj)
-    = TH.FamilyD TH.TypeFam (toName h) (toTyVars h) (toMaybeKind sig)
-  toDec (Exts.DataFamDecl _ _ h sig)
-    = TH.FamilyD TH.DataFam (toName h) (toTyVars h) (toMaybeKind sig)
-#endif
 
   toDec _a@(Exts.FunBind _ mtchs)                           = hsMatchesToFunD mtchs
   toDec (Exts.PatBind _ p rhs bnds)                      = TH.ValD (toPat p)
@@ -613,18 +584,11 @@ instance ToDec (Exts.Decl l) where
   -- the 'vars' bit seems to be for: instance forall a. C (T a) where ...
   -- TH's own parser seems to flat-out ignore them, and honestly I can't see
   -- that it's obviously wrong to do so.
-#if MIN_VERSION_template_haskell(2,11,0)
   toDec (Exts.InstDecl _ Nothing irule ids) = TH.InstanceD
     Nothing
     (toCxt irule)
     (toType irule)
     (toDecs ids)
-#else
-  toDec (Exts.InstDecl _ Nothing irule ids) = TH.InstanceD
-    (toCxt irule)
-    (toType irule)
-    (toDecs ids)
-#endif
 
   toDec (Exts.ClassDecl _ cxt h fds decls) = TH.ClassD
     (toCxt cxt)
@@ -635,6 +599,15 @@ instance ToDec (Exts.Decl l) where
    where
     toFunDep (Exts.FunDep _ ls rs) = TH.FunDep (fmap toName ls) (fmap toName rs)
 
+  toDec (Exts.AnnPragma _ ann) = TH.PragmaD (TH.AnnP (target ann) (expann ann))
+    where
+      target (Exts.Ann _ n _)     = TH.ValueAnnotation (toName n)
+      target (Exts.TypeAnn _ n _) = TH.TypeAnnotation (toName n)
+      target (Exts.ModuleAnn _ _) = TH.ModuleAnnotation
+      expann (Exts.Ann _ _ e)     = toExp e
+      expann (Exts.TypeAnn _ _ e) = toExp e
+      expann (Exts.ModuleAnn _ e) = toExp e
+
   toDec x = todo "toDec" x
 
 instance ToMaybeKind (Exts.ResultSig l) where
@@ -645,10 +618,8 @@ instance ToMaybeKind a => ToMaybeKind (Maybe a) where
     toMaybeKind Nothing  = Nothing
     toMaybeKind (Just a) = toMaybeKind a
 
-#if MIN_VERSION_template_haskell(2,11,0)
 instance ToInjectivityAnn (Exts.InjectivityInfo l) where
   toInjectivityAnn (Exts.InjectivityInfo _ n ns) = TH.InjectivityAnn (toName n) (fmap toName ns)
-#endif
 
 transAct :: Maybe (Exts.Activation l) -> TH.Phases
 transAct Nothing                       = TH.AllPhases
@@ -713,7 +684,7 @@ instance ToType (Exts.InstHead l) where
 
 qualConDeclToCon :: Exts.QualConDecl l -> TH.Con
 qualConDeclToCon (Exts.QualConDecl _ Nothing Nothing cdecl) = conDeclToCon cdecl
-qualConDeclToCon (Exts.QualConDecl _ ns cxt cdecl) = TH.ForallC (toTyVars ns)
+qualConDeclToCon (Exts.QualConDecl _ ns cxt cdecl) = TH.ForallC (toTyVarSpec <$> toTyVars ns)
                                                     (toCxt cxt)
                                                     (conDeclToCon cdecl)
 
@@ -839,5 +810,3 @@ instance ToDecs (Exts.Decl l) where
 
 instance ToDecs a => ToDecs [a] where
   toDecs a = concatMap toDecs a
-
------------------------------------------------------------------------------
